#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

from typing import Optional, Set

import gymnasium as gym

from upkie.utils.robot_state import RobotState

from .upkie_servos import UpkieServos


class UpkieServoPositions(UpkieServos):
    r"""!
    Command servos by position control.

    \anchor upkie_servo_positions_description

    ### Action space

    The action space is consists of the following targets for each joint:

    - `position`: commanded joint angle \f$\theta^*\f$ in [rad] (NaN to
       disable) (required).
    - `kp_scale`: scaling factor \f$k_{p}^{\mathit{scale}}\f$ applied to the
       position feedback gain, between zero and one.
    - `kd_scale`: scaling factor \f$k_{d}^{\mathit{scale}}\f$ applied to the
       velocity feedback gain, between zero and one.

    This makes the agent command servo positions with velocity damping:

    \f[
    \begin{align*}
    \tau & = k_{p} k_{p}^{\mathit{scale}} (\theta^* - \theta) - k_{d}
    k_{d}^{\mathit{scale}} \dot{\theta}
    \end{align*}
    \f]

    ### Observation space

    This environment has the same observation space as
    [UpkieServos](\ref upkie_servos_description).
    """

    ACTION_MASK: Set[str] = set(["position", "kp_scale", "kd_scale"])

    ## \var action_space
    ## Action space.
    action_space: gym.spaces.dict.Dict

    def __init__(
        self,
        frequency: float = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        regulate_frequency: bool = True,
        shm_name: str = "/upkie",
        spine_config: Optional[dict] = None,
    ):
        r"""!
        Initialize environment.

        \param frequency Regulated frequency of the control loop, in Hz.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param regulate_frequency Enables loop frequency regulation.
        \param shm_name Name of shared-memory file.
        \param spine_config Additional spine configuration overriding the
            default `upkie.config.SPINE_CONFIG`. The combined configuration
            dictionary is sent to the spine at every reset.
        """
        super().__init__(
            frequency=frequency,
            frequency_checks=frequency_checks,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )
        action_space = {
            joint.name: gym.spaces.Dict(
                {
                    "position": gym.spaces.Box(
                        low=joint.limit.lower,
                        high=joint.limit.upper,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kp_scale": gym.spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kd_scale": gym.spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            for joint in self.model.joints
        }
        self.action_space = gym.spaces.Dict(action_space)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Any, Optional, Set, Tuple

import gymnasium as gym
import numpy as np
import upkie_description
from loop_rate_limiters import RateLimiter

import upkie.config
from upkie.exceptions import UpkieException
from upkie.model import Model
from upkie.spine import SpineInterface
from upkie.utils.clamp import clamp_and_warn
from upkie.utils.nested_update import nested_update
from upkie.utils.robot_state import RobotState


class UpkieServos(gym.Env):
    r"""!
    Base Upkie environment where actions command servomotors directly.

    \anchor upkie_servos_description

    Actions and observations correspond to the moteus servo API. Under the
    hood, the environment provides a number of features:

    - Communication with the spine process.
    - Initial state randomization (e.g. when training a policy).
    - Loop frequency regulation (optional).

    Note that Upkie environments are made to run on a single CPU thread. The
    downside for reinforcement learning is that computations are not massively
    parallel. The upside is that it simplifies deployment to the real robot, as
    it relies on the same spine interface that runs on real robots.

    ### Action space

    The action space is a dictionary with one key for each servo:

    - `left_hip`: left hip joint (qdd100)
    - `left_knee`: left knee joint (qdd100)
    - `left_wheel`: left wheel joint (mj5208)
    - `right_hip`: right hip joint (qdd100)
    - `right_knee`: right knee joint (qdd100)
    - `right_wheel`: right wheel joint (mj5208)

    The value for each servo dictionary is itself a dictionary with the
    following keys:

    - `position`: commanded joint angle \f$\theta^*\f$ in [rad] (NaN to
       disable) (required).
    - `velocity`: commanded joint velocity \f$\dot{\theta}^*\f$ in [rad] /
       [s] (required).
    - `feedforward_torque`: feedforward joint torque \f$\tau_{\mathit{ff}}\f$
       in [N m].
    - `kp_scale`: scaling factor \f$k_{p}^{\mathit{scale}}\f$ applied to the
       position feedback gain, between zero and one.
    - `kd_scale`: scaling factor \f$k_{d}^{\mathit{scale}}\f$ applied to the
       velocity feedback gain, between zero and one.
    - `maximum_torque`: maximum joint torque \f$\tau_{\mathit{max}}\f$
       (feedforward + feedback) enforced during the whole actuation step, in
       [N m].

    The resulting torque applied by the servo is then:

    \f[
    \begin{align*}
    \tau & = \underset{
            [-\tau_{\mathit{max}}, +\tau_{\mathit{max}}]}{
            \mathrm{clamp}
        }
        \left(
            \tau_{\mathit{ff}} +
            k_{p} k_{p}^{\mathit{scale}} (\theta^* - \theta) +
            k_{d} k_{d}^{\mathit{scale}} (\dot{\theta}^* - \dot{\theta}))
        \right)
    \end{align*}
    \f]

    Position and velocity gains \f$k_{p}\f$ and \f$k_{d}\f$ are configured in
    each moteus controller directly and don't change during execution. We can
    rather modulate the overall feedback gains via the normalized parameters
    \f$k_{p}^{\mathit{scale}} \in [0, 1]\f$ and \f$k_{d}^{\mathit{scale}} \in
    [0, 1]\f$. Note that the servo regulates the torque above at its own
    frequency, which is higher (typically 40 kHz) than the agent and the spine
    frequencies. See the [moteus
    reference](https://github.com/mjbots/moteus/blob/13c171c697ce6f60a73c9385e6fe951957313d1d/docs/reference.md#theory-of-operation)
    for more details.

    ### Observation space

    The observation space is a dictionary with one key for each servo. The
    value for each key is a dictionary with keys:

    - `position`: Joint angle in [rad].
    - `velocity`: Joint velocity in [rad] / [s].
    - `torque`: Joint torque in [N m].
    - `temperature`: Servo temperature in degree Celsius.
    - `voltage`: Power bus voltage of the servo, in [V].

    As with all Upkie environments, full observations from the spine (detailed
    in \ref observations) are also available in the `info` dictionary
    returned by the reset and step functions.
    """

    __bonus_action: dict
    __frequency: Optional[float]
    __rate: Optional[RateLimiter]
    __regulate_frequency: bool
    _spine: SpineInterface
    _spine_config: dict

    ACTION_KEYS: Tuple[str, str, str, str, str, str] = (
        "position",
        "velocity",
        "feedforward_torque",
        "kp_scale",
        "kd_scale",
        "maximum_torque",
    )

    ACTION_MASK: Set[str] = set()

    ## \var action_space
    ## Action space.
    action_space: gym.spaces.dict.Dict

    ## \var init_state
    ## Initial state for the floating base of the robot, which may be
    ## randomized upon resets.
    init_state: RobotState

    ## \var model
    ## Robot model read from its URDF description.
    model: Model

    ## \var observation_space
    ## Observation space.
    observation_space: gym.spaces.dict.Dict

    ## \var version
    ## Environment version number.
    version = 5

    def __init__(
        self,
        frequency: Optional[float] = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        regulate_frequency: bool = True,
        shm_name: str = "/upkie",
        spine_config: Optional[dict] = None,
    ) -> None:
        r"""!
        Initialize environment.

        \param frequency Regulated frequency of the control loop, in Hz. Can be
            prescribed even when `regulate_frequency` is unset, in which case
            `self.dt` will be defined but the loop frequency will not be
            regulated.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param regulate_frequency If set (default), the environment will
            regulate the control loop frequency to the value prescribed in
            `frequency`.
        \param shm_name Name of shared-memory file to exchange with the spine.
        \param spine_config Additional spine configuration overriding the
            default `upkie.config.SPINE_CONFIG`. The combined configuration
            dictionary is sent to the spine at every reset.

        \throw SpineError If the spine did not respond after the prescribed
            number of trials.
        """
        merged_spine_config = upkie.config.SPINE_CONFIG.copy()
        if spine_config is not None:
            nested_update(merged_spine_config, spine_config)
        if regulate_frequency and frequency is None:
            raise UpkieException(f"{regulate_frequency=} but {frequency=}")
        if init_state is None:
            init_state = RobotState(
                position_base_in_world=np.array([0.0, 0.0, 0.6])
            )

        action_space = {}
        neutral_action = {}
        max_action = {}
        min_action = {}
        servo_space = {}

        model = Model(upkie_description.URDF_PATH)
        for joint in model.joints:
            action_space[joint.name] = gym.spaces.Dict(
                {
                    "position": gym.spaces.Box(
                        low=joint.limit.lower,
                        high=joint.limit.upper,
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": gym.spaces.Box(
                        low=-joint.limit.velocity,
                        high=+joint.limit.velocity,
                        shape=(1,),
                        dtype=float,
                    ),
                    "feedforward_torque": gym.spaces.Box(
                        low=-joint.limit.effort,
                        high=+joint.limit.effort,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kp_scale": gym.spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kd_scale": gym.spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "maximum_torque": gym.spaces.Box(
                        low=0.0,
                        high=joint.limit.effort,
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            servo_space[joint.name] = gym.spaces.Dict(
                {
                    "position": gym.spaces.Box(
                        low=joint.limit.lower,
                        high=joint.limit.upper,
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": gym.spaces.Box(
                        low=-joint.limit.velocity,
                        high=+joint.limit.velocity,
                        shape=(1,),
                        dtype=float,
                    ),
                    "torque": gym.spaces.Box(
                        low=-joint.limit.effort,
                        high=+joint.limit.effort,
                        shape=(1,),
                        dtype=float,
                    ),
                    "temperature": gym.spaces.Box(
                        low=0.0,
                        high=100.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "voltage": gym.spaces.Box(
                        low=10.0,  # moteus min 10 V
                        high=44.0,  # moteus max 44 V
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            neutral_action[joint.name] = {
                "position": np.nan,
                "velocity": 0.0,
                "feedforward_torque": 0.0,
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": joint.limit.effort,
            }
            max_action[joint.name] = {
                "position": joint.limit.upper,
                "velocity": joint.limit.velocity,
                "feedforward_torque": joint.limit.effort,
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": joint.limit.effort,
            }
            min_action[joint.name] = {
                "position": joint.limit.lower,
                "velocity": -joint.limit.velocity,
                "feedforward_torque": -joint.limit.effort,
                "kp_scale": 0.0,
                "kd_scale": 0.0,
                "maximum_torque": 0.0,
            }

        # gymnasium.Env: action_space
        self.action_space = gym.spaces.Dict(action_space)

        # gymnasium.Env: observation_space
        self.observation_space = gym.spaces.Dict(servo_space)

        # Class attributes
        self.__bonus_action = {"bullet": {}, "log": {}}
        self.__frequency = frequency
        self.__frequency_checks = frequency_checks
        self.__max_action = max_action
        self.__min_action = min_action
        self.__neutral_action = neutral_action
        self.__rate = None
        self.__regulate_frequency = regulate_frequency
        self._spine = SpineInterface(shm_name, retries=10)
        self._spine_config = merged_spine_config
        self.init_state = init_state
        self.model = model

    def __del__(self):
        """!
        Stop the spine when deleting the environment instance.
        """
        self.close()  # we may close twice but that's OK

    def close(self) -> None:
        """!
        Stop the spine properly.
        """
        if hasattr(self, "_spine"):  # in case SpineError was raised in ctor
            self._spine.stop()

    @property
    def dt(self) -> Optional[float]:
        """!
        Regulated period of the control loop in seconds, or `None` if there
        is no loop frequency regulation.
        """
        return 1.0 / self.__frequency if self.__frequency is not None else None

    @property
    def frequency(self) -> Optional[float]:
        """!
        Regulated frequency of the control loop in Hz, or `None` if there is
        no loop frequency regulation.
        """
        return self.__frequency

    def get_neutral_action(self) -> dict:
        r"""!
        Get the neutral action where servos don't move.

        \return Neutral action where servos don't move.
        """
        return self.__neutral_action.copy()

    def update_init_rand(self, **kwargs) -> None:
        r"""!
        Update initial-state randomization.

        Keyword arguments are forwarded as is to \ref
        upkie.utils.robot_state_randomization.RobotStateRandomization.update.
        """
        self.init_state.randomization.update(**kwargs)

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[np.ndarray, dict]:
        r"""!
        Resets the spine and get an initial observation.

        \param seed Number used to initialize the environment’s internal random
            number generator.
        \param options Currently unused.
        \return
            - `observation`: Initial vectorized observation, i.e. an element
              of the environment's `observation_space`.
            - `info`: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        super().reset(seed=seed)
        self._spine.stop()
        self.__reset_rate()
        self.__reset_init_state()
        spine_observation = self._spine.start(self._spine_config)
        observation = self.__get_observation(spine_observation)
        info = {"spine_observation": spine_observation}
        return observation, info

    def __reset_rate(self):
        if self.__regulate_frequency:
            rate_name = f"{self.__class__.__name__} rate limiter"
            self.__rate = RateLimiter(
                self.__frequency,
                name=rate_name,
                warn=self.__frequency_checks,
            )

    def __reset_init_state(self):
        init_state, np_random = self.init_state, self.np_random
        orientation_matrix = init_state.sample_orientation(np_random)
        qx, qy, qz, qw = orientation_matrix.as_quat()
        orientation_quat = np.array([qw, qx, qy, qz])
        position = init_state.sample_position(np_random)
        linear_velocity = init_state.sample_linear_velocity(np_random)
        omega = init_state.sample_angular_velocity(np_random)

        bullet_config = self._spine_config["bullet"]
        reset = bullet_config["reset"]
        reset["orientation_base_in_world"] = orientation_quat
        reset["position_base_in_world"] = position
        reset["linear_velocity_base_to_world_in_world"] = linear_velocity
        reset["angular_velocity_base_in_base"] = omega
        reset["joint_configuration"] = init_state.joint_configuration

    def step(
        self,
        action: np.ndarray,
    ) -> Tuple[np.ndarray, float, bool, bool, dict]:
        r"""!
        Run one timestep of the environment's dynamics.

        When the end of the episode is reached, you are responsible for calling
        `reset()` to reset the environment's state.

        \param action Action from the agent.
        \return
            - `observation`: Observation of the environment, i.e. an element
              of its `observation_space`.
            - `reward`: Reward returned after taking the action.
            - `terminated`: Whether the agent reached a terminal state,
              which may be a good or a bad thing. When true, the user needs to
              call `reset()`.
            - `truncated`: Whether the episode is reaching max number of
              steps. This boolean can signal a premature end of the episode,
              i.e. before a terminal state is reached. When true, the user
              needs to call `reset()`.
            - `info`: Dictionary with additional information, reporting in
              particular the full observation dictionary coming from the spine.
        """
        if self.__regulate_frequency:
            self.__rate.sleep()  # wait until clock tick to send the action
            self.log("rate", {"slack": self.__rate.slack})

        # Prepare spine action
        spine_action = self.__get_spine_action(action)
        for key in ("bullet", "log"):
            if not self.__bonus_action[key]:
                continue
            spine_action[key] = {}
            spine_action[key].update(self.__bonus_action[key])
            self.__bonus_action[key].clear()

        # Send action to and get observation from the spine
        spine_observation = self._spine.set_action(spine_action)

        # Process spine observation
        observation = self.__get_observation(spine_observation)
        reward = 1.0  # ready for e.g. an ObservationBasedReward wrapper
        terminated = False
        truncated = False  # will be handled by e.g. a TimeLimit wrapper
        info = {"spine_observation": spine_observation}
        return observation, reward, terminated, truncated, info

    def __get_observation(self, spine_observation: dict):
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Full observation dictionary from the spine.
        \return Environment observation.
        """
        # If creating a new object turns out to be too slow we can switch to
        # updating in-place.
        return {
            joint.name: {
                key: np.array(
                    [spine_observation["servo"][joint.name][key]],
                    dtype=float,
                )
                for key in self.observation_space[joint.name]
            }
            for joint in self.model.joints
        }

    def __get_spine_action(self, env_action: dict) -> dict:
        r"""!
        Convert environment action to a spine action dictionary.

        \param env_action Environment action.
        \return Spine action dictionary.
        """
        spine_action = {"servo": {}}
        for joint in self.model.joints:
            servo_action = {}
            for key in self.ACTION_KEYS:
                action = (
                    env_action[joint.name][key]
                    if key in env_action[joint.name]
                    and (not self.ACTION_MASK or key in self.ACTION_MASK)
                    else self.__neutral_action[joint.name][key]
                )
                action_value = (
                    action.item()
                    if isinstance(action, np.ndarray)
                    else float(action)
                )
                servo_action[key] = clamp_and_warn(
                    action_value,
                    self.__min_action[joint.name][key],
                    self.__max_action[joint.name][key],
                    label=f"{joint.name}: {key}",
                )
            spine_action["servo"][joint.name] = servo_action
        return spine_action

    def log(self, name: str, entry: Any) -> None:
        r"""!
        Log a new entry to the "log" key of the action dictionary.

        \param name Name of the entry.
        \param entry Dictionary to log along with the actual action.
        """
        if isinstance(entry, dict):
            self.__bonus_action["log"][name] = entry.copy()
        else:  # logging values directly
            self.__bonus_action["log"][name] = entry

    def get_bullet_action(self) -> dict:
        r"""!
        Get the Bullet action that will be applied at next step.

        \return Upcoming simulator action.
        """
        return self.__bonus_action["bullet"]

    def set_bullet_action(self, bullet_action: dict) -> None:
        r"""!
        Prepare for the next step an extra action for the Bullet spine.

        This extra action can be for instance a set of external forces applied
        to some robot bodies.

        \param bullet_action Action dictionary processed by the Bullet spine.
        """
        self.__bonus_action["bullet"] = bullet_action.copy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""!
Inter-process communication interface.
"""

import mmap
import sys
from time import perf_counter_ns

import msgpack

from ..exceptions import PerformanceIssue, SpineError
from .request import Request
from .serialize import serialize
from .wait_for_shared_memory import wait_for_shared_memory


class SpineInterface:
    """!
    Interface to interact with a spine from a Python agent.
    """

    _mmap: mmap.mmap

    def __init__(
        self,
        shm_name: str = "/upkie",
        retries: int = 1,
        perf_checks: bool = True,
    ):
        r"""!
        Connect to the spine shared memory.

        \param shm_name Name of the shared memory object.
        \param retries Number of times to try opening the shared-memory file.
        \param perf_checks If true, run performance checks after construction.
        """
        shared_memory = wait_for_shared_memory(shm_name, retries)
        self._mmap = shared_memory._mmap
        self._packer = msgpack.Packer(default=serialize, use_bin_type=True)
        self._shared_memory = shared_memory
        self._stop_waiting = set([Request.kNone, Request.kError])
        self._unpacker = msgpack.Unpacker(raw=False)
        if perf_checks:
            self.__perf_checks()

    def __perf_checks(self):
        packer_cls = str(msgpack.Packer)
        unpacker_cls = str(msgpack.Unpacker)
        if "fallback" in packer_cls or "fallback" in unpacker_cls:
            # See https://github.com/upkie/upkie/issues/377
            raise PerformanceIssue("msgpack is running in pure Python")

    def __del__(self):
        """!
        Close memory mapping.

        Note that the spine process will unlink the shared memory object, so we
        don't unlink it here.
        """
        if hasattr(self, "_shared_memory"):  # handle ctor exceptions
            self._shared_memory.close()

    def set_action(self, action: dict) -> dict:
        r"""!
        Set action for the spine to process.

        \param[in] action Action dictionary.
        \return Observation dictionary.
        """
        self._wait_for_spine()
        self._write_dict(action)
        self._write_request(Request.kAction)
        self._wait_for_spine()
        observation = self._read_dict()
        return observation

    def start(self, config: dict) -> dict:
        r"""!
        Reset the spine to a new configuration.

        \param[in] config Configuration dictionary.
        \return Observation dictionary.
        """
        self._wait_for_spine()
        self._write_dict(config)
        self._write_request(Request.kStart)
        self._wait_for_spine()
        observation = self._read_dict()
        return observation

    def stop(self) -> None:
        """!
        Tell the spine to stop all actuators.
        """
        self._wait_for_spine()
        self._write_request(Request.kStop)

    def _read_request(self) -> int:
        """!
        Read current request from shared memory.
        """
        self._mmap.seek(0)
        return int.from_bytes(self._mmap.read(4), byteorder=sys.byteorder)

    def _read_dict(self) -> dict:
        r"""!
        Read dictionary from shared memory.

        \return Observation dictionary.
        """
        assert self._read_request() == Request.kNone
        self._mmap.seek(0)
        self._mmap.read(4)  # skip request field
        size = int.from_bytes(self._mmap.read(4), byteorder=sys.byteorder)
        data = self._mmap.read(size)
        self._unpacker.feed(data)
        unpacked = 0
        last_dict = {}
        for observation in self._unpacker:
            last_dict = observation
            unpacked += 1
        assert unpacked == 1
        return last_dict

    def _wait_for_spine(self, timeout_ns: int = 100000000) -> None:
        r"""!
        Wait for the spine to signal itself as available, which it does by
        setting the current request to none in shared memory.

        \param timeout_ns Don't wait for more than this duration in
            nanoseconds.
        \raise SpineError If the request read from spine has an error flag.
        """
        stop = perf_counter_ns() + timeout_ns
        while self._read_request() not in self._stop_waiting:  # sets are fast
            # Fun fact: `not in set` is 3-4x faster than `!=` on the raspi
            # perf_counter_ns clocks ~1 us on the raspi
            if perf_counter_ns() > stop:
                raise TimeoutError(
                    "Spine did not process request within "
                    f"{timeout_ns / 1e6:.1f} ms, is it stopped?"
                )
        if self._read_request() == Request.kError:
            self._write_request(Request.kNone)
            raise SpineError("Invalid request, is the spine started?")

    def _write_request(self, request: int) -> None:
        """!
        Set request in shared memory.
        """
        self._mmap.seek(0)
        self._mmap.write(request.to_bytes(4, byteorder=sys.byteorder))

    def _write_dict(self, dictionary: dict) -> None:
        r"""!
        Set the shared memory to a given dictionary.

        \param dictionary Dictionary to pack and write.
        """
        assert self._read_request() == Request.kNone
        data = self._packer.pack(dictionary)
        size = len(data)
        self._mmap.seek(0)
        self._mmap.read(4)  # skip request field
        self._mmap.write(size.to_bytes(4, byteorder=sys.byteorder))
        self._mmap.write(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Add noise to the action of an environment."""

import gymnasium as gym
import numpy as np

from upkie.exceptions import UpkieException


class NoisifyAction(gym.ActionWrapper):
    """!
    Add noise to the action of an environment.
    """

    ## \var high
    ## Upper bound on action noise.
    high: np.ndarray

    ## \var low
    ## Lower bound on action noise.
    low: np.ndarray

    def __init__(self, env, noise: np.ndarray):
        r"""!
        Create wrapper.

        \param env Environment to wrap.
        \param noise Noise level.
        """
        super().__init__(env)
        if noise.shape != env.action_space.shape:
            raise UpkieException(
                f"Action {noise.shape=} does not "
                f"match {env.action_space.shape=}"
            )
        self.high = +np.abs(noise)
        self.low = -np.abs(noise)

    def action(self, action: np.ndarray) -> np.ndarray:
        r"""!
        Get the noisy action.

        \param action Original action.
        \return Noisy action.
        """
        noise = self.np_random.uniform(low=self.low, high=self.high)
        noisy_action = np.clip(
            action + noise,
            self.env.action_space.low,
            self.env.action_space.high,
        )
        return noisy_action



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

## \namespace upkie.config
## \brief Static robot configuration.

from .spine_config import SPINE_CONFIG

__all__ = [
    "SPINE_CONFIG",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""!
Domain randomization of robot states.
"""

from typing import Optional

import numpy as np
from scipy.spatial.transform import Rotation as ScipyRotation


class RobotStateRandomization:
    r"""!
    Domain randomization parameters for Upkie's state.

    \note The API for this class is likely to change. Initially all its
    attributes where floating-point numbers due to a quick-and-dirty design
    decision.
    """

    ## \var roll
    ## Amount of roll angle randomization, in radians.
    roll: float

    ## \var pitch
    ## Amount of pitch angle randomization, in radians.
    pitch: float

    ## \var x
    ## Amount of x-axis position randomization, in meters.
    x: float

    ## \var z
    ## Amount of z-axis position randomization, in meters.
    z: float

    ## \var omega_x
    ## Amount of x-axis randomization on the angular-velocity vector, in rad/s.
    omega_x: float

    ## \var omega_y
    ## Amount of y-axis randomization on the angular-velocity vector, in rad/s.
    omega_y: float

    ## \var linear_velocity
    ## Magnitude of linear velocity randomization, as a 3D vector in m/s.
    linear_velocity: np.ndarray

    def __init__(
        self,
        roll: float = 0.0,
        pitch: float = 0.0,
        x: float = 0.0,
        z: float = 0.0,
        omega_x: float = 0.0,
        omega_y: float = 0.0,
        linear_velocity: Optional[np.ndarray] = None,
    ):
        r"""!
        Initialize sampler.

        \param roll Amount of roll angle randomization, in radians.
        \param pitch Amount of pitch angle randomization, in radians.
        \param x Amount of x-axis position randomization, in meters.
        \param z Amount of z-axis position randomization, in meters.
        \param omega_x Amount of x-axis randomization on the angular-velocity
            vector, in rad/s.
        \param omega_y Amount of y-axis randomization on the angular-velocity
            vector, in rad/s.
        \param linear_velocity Magnitude of linear velocity randomization, as a
            3D vector in m/s.
        """
        self.roll = roll
        self.pitch = pitch
        self.x = x
        self.z = z
        self.omega_x = omega_x
        self.omega_y = omega_y
        self.linear_velocity = (
            linear_velocity if linear_velocity is not None else np.zeros(3)
        )

    def update(
        self,
        roll: Optional[float] = None,
        pitch: Optional[float] = None,
        x: Optional[float] = None,
        z: Optional[float] = None,
        omega_x: Optional[float] = None,
        omega_y: Optional[float] = None,
        v_x: Optional[float] = None,
        v_z: Optional[float] = None,
    ) -> None:
        r"""!
        Update some fields.

        \param roll Roll angle for the rotation from the randomized frame to
            the parent frame (Euler Z-Y-X convention).
        \param pitch Pitch angle for the rotation from the randomized frame to
            the parent frame (Euler Z-Y-X convention).
        \param x Translation of the randomized frame along the x-axis of the
            parent frame.
        \param z Translation of the randomized frame along the z-axis of the
            parent frame.
        \param omega_x Angular velocity from the randomized frame to the parent
            frame, expressed in the randomized frame, along the x-axis.
        \param omega_y Angular velocity from the randomized frame to the parent
            frame, expressed in the randomized frame, along the y-axis.
        \param v_x Linear velocity from the randomized frame to the parent
            frame, expressed in the parent frame and along the x-axis.
        \param v_z Linear velocity from the randomized frame to the parent
            frame, expressed in the parent frame and along the z-axis.
        """
        if roll is not None:
            self.roll = roll
        if pitch is not None:
            self.pitch = pitch
        if x is not None:
            self.x = x
        if z is not None:
            self.z = z
        if omega_x is not None:
            self.omega_x = omega_x
        if omega_y is not None:
            self.omega_y = omega_y
        if v_x is not None:
            self.linear_velocity[0] = v_x
        if v_z is not None:
            self.linear_velocity[2] = v_z

    def sample_orientation(
        self, np_random: np.random.Generator
    ) -> ScipyRotation:
        r"""!
        Sample an orientation within the given bounds.

        \param[in] np_random NumPy random number generator.
        \return Sampled rotation matrix.
        """
        yaw_pitch_roll_bounds = np.array([0.0, self.pitch, self.roll])
        yaw_pitch_roll = np_random.uniform(
            low=-yaw_pitch_roll_bounds,
            high=+yaw_pitch_roll_bounds,
            size=3,
        )
        return ScipyRotation.from_euler("ZYX", yaw_pitch_roll)

    def sample_position(self, np_random: np.random.Generator) -> np.ndarray:
        r"""!
        Sample a position within the given bounds.

        \param[in] np_random NumPy random number generator.
        \return Sampled position vector.
        """
        return np_random.uniform(
            low=np.array([-self.x, 0.0, 0.0]),
            high=np.array([+self.x, 0.0, self.z]),
            size=3,
        )

    def sample_angular_velocity(
        self, np_random: np.random.Generator
    ) -> np.ndarray:
        r"""!
        Sample an angular velocity within the given bounds.

        \param[in] np_random NumPy random number generator.
        \return Sampled angular-velocity vector.
        """
        return np_random.uniform(
            low=np.array([-self.omega_x, -self.omega_y, 0.0]),
            high=np.array([+self.omega_x, +self.omega_y, 0.0]),
            size=3,
        )

    def sample_linear_velocity(
        self, np_random: np.random.Generator
    ) -> np.ndarray:
        r"""!
        Sample a linear velocity within the given bounds.

        \param[in] np_random NumPy random number generator.
        \return Sampled linear-velocity vector.
        """
        return np_random.uniform(
            low=-self.linear_velocity,
            high=self.linear_velocity,
            size=3,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

## \namespace upkie.envs.rewards
## \brief Reward functions used in Gymnasium environments.

from .wheeled_inverted_pendulum_reward import WheeledInvertedPendulumReward

__all__ = [
    "WheeledInvertedPendulumReward",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""!
Robot state with optional randomization.
"""

from typing import Optional

import numpy as np
from scipy.spatial.transform import Rotation as ScipyRotation

from upkie.utils.robot_state_randomization import RobotStateRandomization


class RobotState:
    """!
    Robot state (configuration and velocity) with optional randomization.
    """

    ## \var angular_velocity_base_in_base
    ## Angular velocity of the base in body coordinates, in rad/s.
    angular_velocity_base_in_base: np.ndarray

    ## \var joint_configuration
    ## Vector of joint angles in radians.
    joint_configuration: np.ndarray

    ## \var joint_velocity
    ## Vector of joint velocities, in rad/s.
    joint_velocity: np.ndarray

    ## \var linear_velocity_base_to_world_in_world
    ## Linear velocity of the base in world coordinates, in m/s.
    linear_velocity_base_to_world_in_world: np.ndarray

    ## \var orientation_base_in_world
    ## Orientation of the base frame with respect to the world frame.
    orientation_base_in_world: ScipyRotation

    ## \var position_base_in_world
    ## Position of the base frame in the world frame.
    position_base_in_world: np.ndarray

    ## \var randomization
    ## Displacements and velocities added to the state when calling one of the
    ## sampling functions.
    randomization: RobotStateRandomization

    def __init__(
        self,
        angular_velocity_base_in_base: Optional[np.ndarray] = None,
        joint_configuration: Optional[np.ndarray] = None,
        joint_velocity: Optional[np.ndarray] = None,
        linear_velocity_base_to_world_in_world: Optional[np.ndarray] = None,
        orientation_base_in_world: Optional[ScipyRotation] = None,
        position_base_in_world: Optional[np.ndarray] = None,
        randomization: Optional[RobotStateRandomization] = None,
    ):
        r"""!
        Initialize a new state.

        \param[in] angular_velocity_base_in_base Body angular velocity of the
            floating-base link.
        \param[in] joint_configuration Joint configuration vector (not
            including the floating-base joint).
        \param[in] joint_velocity Joint velocity vector (not including the
            floating-base joint).
        \param[in] linear_velocity_base_to_world_in_world Linear velocity of
            the floating-base link in the world frame.
        \param[in] orientation_base_in_world Rotation from the floating-base
            frame to the world frame.
        \param[in] position_base_in_world Position of the floating-base frame
            in the world frame.
        \param[in] randomization Optional state randomization distribution.
        """
        self.angular_velocity_base_in_base = (
            angular_velocity_base_in_base
            if angular_velocity_base_in_base is not None
            else np.zeros(3)
        )
        self.joint_configuration = (
            joint_configuration
            if joint_configuration is not None
            else np.zeros(6)
        )
        self.joint_velocity = (
            joint_velocity if joint_velocity is not None else np.zeros(6)
        )
        self.linear_velocity_base_to_world_in_world = (
            linear_velocity_base_to_world_in_world
            if linear_velocity_base_to_world_in_world is not None
            else np.zeros(3)
        )
        self.orientation_base_in_world = (
            orientation_base_in_world
            if orientation_base_in_world is not None
            else ScipyRotation.identity()
        )
        self.position_base_in_world = (
            position_base_in_world
            if position_base_in_world is not None
            else np.array([0.0, 0.0, 0.6])  # Upkie above horizontal plane
        )
        self.randomization = (
            randomization
            if randomization is not None
            else RobotStateRandomization()
        )

    def sample_angular_velocity(
        self, np_random: np.random.Generator
    ) -> np.ndarray:
        r"""!
        Sample an angular velocity around the one in this state.

        \param[in] np_random NumPy random number generator.
        \return Sampled angular velocity.
        """
        angular_velocity_rand_to_base_in_base = (
            self.randomization.sample_angular_velocity(np_random)
        )
        angular_velocity_base_to_world_in_base = (
            self.angular_velocity_base_in_base
        )
        return (
            angular_velocity_base_to_world_in_base
            + angular_velocity_rand_to_base_in_base
        )

    def sample_linear_velocity(
        self, np_random: np.random.Generator
    ) -> np.ndarray:
        r"""!
        Sample a linear velocity around the one in this state.

        \param[in] np_random NumPy random number generator.
        \return Sampled linear velocity.
        """
        linear_velocity_rand_to_world_in_world = (
            self.randomization.sample_linear_velocity(np_random)
        )
        return (
            self.linear_velocity_base_to_world_in_world
            + linear_velocity_rand_to_world_in_world
        )

    def sample_orientation(
        self, np_random: np.random.Generator
    ) -> ScipyRotation:
        r"""!
        Sample an orientation around the one in this state.

        \param[in] np_random NumPy random number generator.
        \return Sampled orientation.
        """
        rotation_rand_to_base = self.randomization.sample_orientation(
            np_random
        )
        rotation_base_to_world = self.orientation_base_in_world
        return rotation_base_to_world * rotation_rand_to_base

    def sample_position(self, np_random: np.random.Generator) -> np.ndarray:
        r"""!
        Sample a position around the one in this state.

        \param[in] np_random NumPy random number generator.
        \return Randomized position.
        """
        position_rand_in_world = self.randomization.sample_position(np_random)
        return self.position_base_in_world + position_rand_in_world



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

from .joint_limit import JointLimit


class Joint:
    r"""!
    Joint properties (see also [joint limits](\ref joint-limits)).

    Extra indices in this class are provided as convenience for users familiar
    with Pinocchio.
    """

    ## \var index
    ## Index of the joint in the servo layout. Starts from 1 as in Pinocchio.
    index: int

    ## \var idx_q
    ## Index of the joint in configuration vectors.
    idx_q: int

    ## \var idx_v
    ## Index of the joint in tangent (velocity, torque) vectors.
    idx_v: int

    ## \var name
    ## Name of the joint.
    name: str

    ## \var limit
    ## Joint limit in configuration and tangent spaces.
    limit: JointLimit

    def __init__(
        self,
        index: int,
        idx_q: int,
        idx_v: int,
        name: str,
        limit: JointLimit,
    ):
        """!
        Manual constructor for joint properties.

        (We don't use a `dataclass` because Doxygen does not detect
        attribute-only declarations in Python as of version 1.9.1.)
        """
        self.index = index
        self.idx_q = idx_q
        self.idx_v = idx_v
        self.name = name
        self.limit = limit



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""!
Import `logging` from this module to use logging from Python standard library
with formatting similar to spdlog.
"""

import logging


class SpdlogFormatter(logging.Formatter):
    """!
    Custom logging formatter visually consistent with spdlog.
    """

    def __init__(self):
        """!
        Initialize log formatter.
        """
        BOLD_RED: str = "\033[31;1m"
        BOLD_WHITE: str = "\033[37;1m"
        BOLD_YELLOW: str = "\033[33;1m"
        GREEN: str = "\033[32m"
        ON_RED: str = "\033[41m"
        RESET: str = "\033[0m"

        self.level_format: dict = {
            logging.CRITICAL: f"[{ON_RED}{BOLD_WHITE}critical{RESET}]",
            logging.DEBUG: "[debug]",
            logging.ERROR: f"[{BOLD_RED}error{RESET}]",
            logging.INFO: f"[{GREEN}info{RESET}]",
            logging.WARNING: f"[{BOLD_YELLOW}warning{RESET}]",
        }

    def format(self, record):
        r"""!
        Format a given record.

        \param record Record to format.
        """
        custom_format = (
            "[%(asctime)s] "
            + self.level_format.get(record.levelno, "[???]")
            + " %(message)s (%(filename)s:%(lineno)d)"
        )
        formatter = logging.Formatter(custom_format)
        return formatter.format(record)


logger = logging.getLogger()
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
handler.setFormatter(SpdlogFormatter())
logger.addHandler(handler)
logging.basicConfig(level=logging.INFO)


__all__ = [
    "logging",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

## \namespace upkie.envs
## \brief Gymnasium environments.
##
## See the [list of environments](\ref gym-environments) for more details.

import gymnasium as gym

from .upkie_ground_velocity import UpkieGroundVelocity
from .upkie_servo_positions import UpkieServoPositions
from .upkie_servo_torques import UpkieServoTorques
from .upkie_servos import UpkieServos


def register() -> None:
    """!
    Register Upkie environments with Gymnasium.
    """
    # Environments
    envs = (
        ("UpkieServoPositions", UpkieServoPositions),
        ("UpkieServoTorques", UpkieServoTorques),
        ("UpkieServos", UpkieServos),
    )
    for env_name, env_class in envs:
        gym.envs.registration.register(
            id=f"{env_name}-v{env_class.version}",
            entry_point=f"upkie.envs:{env_name}",
        )

    # Wrappers
    gym.envs.registration.register(
        id=f"UpkieGroundVelocity-v{UpkieGroundVelocity.version}",
        entry_point=(
            "upkie.envs.upkie_ground_velocity:make_upkie_ground_velocity"
        ),
    )


__all__ = [
    "UpkieGroundVelocity",
    "UpkieServoPositions",
    "UpkieServoTorques",
    "UpkieServos",
    "register",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import gymnasium as gym
import numpy as np

from upkie.exceptions import UpkieException


class NoisifyObservation(gym.ObservationWrapper):
    """!
    Add noise to the observation of an environment.
    """

    ## \var high
    ## Upper bound on observation noise.
    high: np.ndarray

    ## \var low
    ## Lower bound on observation noise.
    low: np.ndarray

    def __init__(self, env, noise: np.ndarray):
        r"""!
        Create wrapper.

        \param env Environment to wrap.
        \param noise Noise level.
        """
        super().__init__(env)
        if noise.shape != env.observation_space.shape:
            raise UpkieException(
                f"Observation {noise.shape=} does not "
                f"match {env.observation_space.shape=}"
            )
        self.high = +np.abs(noise)
        self.low = -np.abs(noise)

    def observation(self, observation: np.ndarray) -> np.ndarray:
        r"""!
        Get noisy observation.

        \param observation Noise-less observation.
        \return Noisy observation.
        """
        noise = self.np_random.uniform(low=self.low, high=self.high)
        noisy_observation = np.clip(
            observation + noise,
            self.env.observation_space.low,
            self.env.observation_space.high,
        ).astype(self.env.observation_space.dtype)
        return noisy_observation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

## \namespace upkie.exceptions
## \brief Exceptions raised by the Python library.


class UpkieException(Exception):
    """!
    Base class for exceptions raised by Upkie agents.
    """


class FallDetected(UpkieException):
    """!
    Raised when a fall is detected.
    """


class MissingOptionalDependency(UpkieException):
    """!
    Raised when an optional feature lacks a corresponding optional dependency.
    """


class ModelError(UpkieException):
    """!
    Raised when something is wrong in the robot model.
    """


class UpkieRuntimeError(UpkieException, RuntimeError):
    """!
    Runtime error, for instance an invalid call to a library function.
    """


class PerformanceIssue(UpkieException):
    """!
    Exception raised when a performance issue is detected.
    """


class SpineError(UpkieException):
    """!
    Exception raised when the spine sets an error flag in the request field of
    the shared memory map.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""!
Convert between various rotation representations.
"""

from typing import Tuple

import numpy as np


def rotation_matrix_from_quaternion(
    quat: Tuple[float, float, float, float]
) -> np.ndarray:
    r"""!
    Convert a unit quaternion to the matrix representing the same rotation.

    \param quat Unit quaternion to convert, in `[w, x, y, z]` format.
    \return Rotation matrix corresponding to this quaternion.

    See `Conversion between quaternions and rotation matrices`_.

    .. _`Conversion between quaternions and rotation matrices`:
        https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles#Rotation_matrices
    """
    if abs(np.dot(quat, quat) - 1.0) > 1e-5:
        raise ValueError(f"Quaternion {quat} is not normalized")
    qw, qx, qy, qz = quat
    return np.array(
        [
            [
                1 - 2 * (qy**2 + qz**2),
                2 * (qx * qy - qz * qw),
                2 * (qw * qy + qx * qz),
            ],
            [
                2 * (qx * qy + qz * qw),
                1 - 2 * (qx**2 + qz**2),
                2 * (qy * qz - qx * qw),
            ],
            [
                2 * (qx * qz - qy * qw),
                2 * (qy * qz + qx * qw),
                1 - 2 * (qx**2 + qy**2),
            ],
        ]
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria
#
## \namespace upkie.model
## \brief Robot model.

from .model import Model

__all__ = [
    "Model",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

from typing import Any, Dict, SupportsFloat, Tuple

import gymnasium as gym
from gymnasium.core import ActType, ObsType


class ObservationBasedReward(gym.Wrapper[ObsType, ActType, ObsType, ActType]):
    r"""!
    Redefine the reward of an environment as a function of its observation.

    If you would like to redefine the reward that is returned by the base
    environment before passing it to learning code, you can simply inherit from
    this class and overwrite the :meth:`reward` method. See also the
    `RewardWrapper` class defined in Gymnasium.

    Rewards in reinforcement learning are often defined as \f$r(s_t, a_t,
    s_{t+1})\f$. With this wrapper, the reward function is redefined as
    \f$r(o_{t+1})\f$, with two differences:

    - We use the observation rather than the environment state.
    - The new reward is computed based on the post-step observation.

    Using observation rather than state does not have a big impact when
    training in simulation, as simulation sensors can read any part of the full
    simulation state on demand.
    """

    def __init__(self, env: gym.Env[ObsType, ActType]):
        r"""!
        Constructor for the Reward wrapper.

        \param[in] env Environment to be wrapped.
        """
        super().__init__(env)

    def step(
        self, action: ActType
    ) -> Tuple[ObsType, SupportsFloat, bool, bool, Dict[str, Any]]:
        r"""!
        Modifies the :attr:`env` reward using :meth:`self.reward`.
        """
        observation, _, terminated, truncated, info = self.env.step(action)
        return (
            observation,
            self.reward(observation, info),
            terminated,
            truncated,
            info,
        )

    def reward(self, observation: ObsType, info: dict) -> SupportsFloat:
        r"""!
        Returns the new environment reward.

        \param[in] observation Latest observation from the environment.
        \param[in] info Latest info dictionary from the environment.
        \return The modified reward.
        """
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import gymnasium as gym
import numpy as np

from upkie.exceptions import UpkieException


class AddActionToObservation(gym.ObservationWrapper):
    """!
    Append last action vector to the observation vector.

    Can be used in combination with gymnasium.wrappers.FrameStackObservation to
    create histories. (If there were directly a history wrapper in gymnasium we
    wouldn't need this wrapper.)
    """

    ## @var observation_space
    ## Observation space.
    observation_space: gym.spaces.Box

    _last_action: np.ndarray

    def __init__(self, env):
        r"""!
        Initialize wrapper.

        \param env Wrapped environment.

        The initial action is set to zero.
        """
        super().__init__(env)
        if env.observation_space.dtype != env.action_space.dtype:
            raise UpkieException(
                "Not sure which type to pick "
                f"between {env.observation_space.dtype=} "
                f"and {env.action_space.dtype=}"
            )
        low = np.concatenate([env.observation_space.low, env.action_space.low])
        self.observation_space = gym.spaces.Box(
            low=low,
            high=np.concatenate(
                [env.observation_space.high, env.action_space.high]
            ),
            shape=low.shape,
            dtype=env.observation_space.dtype,
        )
        self._last_action = np.zeros(
            env.action_space.shape, dtype=env.action_space.dtype
        )

    def observation(self, observation: np.ndarray) -> np.ndarray:
        r"""!
        rGet wrapped observation.

        \param observation Observation from the wrapped environment.
        \return Wrapped observation.
        """
        return np.concatenate([observation, self._last_action])

    def step(self, action: np.ndarray):
        r"""!
        Step the environment.

        \param action Action from the agent.
        """
        self._last_action = action
        return super().step(action)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

## \var SPINE_CONFIG
## Spine configuration dictionary used as defaults by Gymnasium environments.
SPINE_CONFIG = {
    "bullet": {
        "follower_camera": False,
        "gui": True,
        "reset": {
            "orientation_base_in_world": [1.0, 0.0, 0.0, 0.0],
            "position_base_in_world": [0.0, 0.0, 0.6],
        },
        "torque_control": {
            "kp": 20.0,
            "kd": 1.0,
        },
    },
    "floor_contact": {
        "upper_leg_torque_threshold": 10.0,
    },
    "wheel_contact": {
        "cutoff_period": 0.2,
        "liftoff_inertia": 0.001,
        "min_touchdown_acceleration": 2.0,
        "min_touchdown_torque": 0.015,
        "touchdown_inertia": 0.004,
    },
    "wheel_odometry": {
        "signed_radius": {
            "left_wheel": +0.05,
            "right_wheel": -0.05,
        }
    },
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

## \namespace upkie.spine
## \brief Python interface for agents to interact with a spine.

from .request import Request
from .serialize import serialize
from .spine_interface import SpineInterface

__all__ = [
    "Request",
    "SpineInterface",
    "serialize",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Tuple

import gymnasium as gym
import numpy as np


class DifferentiateAction(gym.Wrapper):
    r"""!
    Act on the derivative of the action.
    """

    ## @var action_penalty
    ## Weight for an additional penalty on the differential action added to the
    ## reward.
    action_penalty: float

    ## @var action_space
    ## Action space.
    action_space: gym.spaces.Box

    def __init__(
        self,
        env,
        min_derivative: np.ndarray,
        max_derivative: np.ndarray,
        action_penalty: float = 0.0,
    ):
        r"""!
        Initialize wrapper.

        \param env Environment to wrap.
        \param min_derivative Lower bound on the derivative of the original
            action.
        \param max_derivative Upper bound on the derivative of the original
            action.
        \param action_penalty Weight for an additional penalty on the
            differential action added to the reward.

        \note We assume the original action lives in a vector space.
        """
        super().__init__(env)
        self.action_space = gym.spaces.Box(
            np.float32(min_derivative),
            np.float32(max_derivative),
            shape=env.action_space.shape,
            dtype=np.float32,
        )
        self._integral = np.zeros(env.action_space.shape)
        self.action_penalty = action_penalty

    def reset(self, **kwargs):
        r"""!
        Reset the environment.

        \param kwargs Keyword arguments forwarded to the wrapped environment.
        """
        self._integral = np.zeros(self.action_space.shape)
        return self.env.reset(**kwargs)

    def step(
        self,
        action: np.ndarray,
    ) -> Tuple[np.ndarray, float, bool, bool, dict]:
        r"""!
        Step the environment.

        \param action Action from the agent.
        \return Tuple with (observation, reward, terminated, truncated,info).
            See \ref upkie.envs.upkie_servos.UpkieServos.step for details.
        """
        self._integral = np.clip(
            self._integral + action * self.env.unwrapped.dt,
            self.env.action_space.low,
            self.env.action_space.high,
        )
        observation, reward, terminated, truncated, info = self.env.step(
            self._integral
        )
        wrapped_reward = reward - self.action_penalty * action.dot(action)
        return observation, wrapped_reward, terminated, truncated, info



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

r"""!
Clamping functions.
"""

from typing import Optional

from .spdlog import logging


def clamp(
    value: float, lower: Optional[float] = None, upper: Optional[float] = None
) -> float:
    r"""!
    Clamp a value between a lower and an upper bound.

    \param value Value to clamp.
    \param lower Lower bound.
    \param upper Upper bound.
    """
    if lower is not None and value < lower:
        return lower
    if upper is not None and value > upper:
        return upper
    return value


def clamp_abs(value: float, bound: float):
    r"""!
    Clamp a value's absolute value, keeping its sign.

    \param value Value to clamp.
    \param bound Absolute value bound.
    """
    return clamp(value, -bound, bound)


def clamp_and_warn(value: float, lower: float, upper: float, label: str):
    r"""!
    Clamp a value between a lower and an upper bound, warning if the value is
    changed.

    \param value Value to clamp.
    \param lower Lower bound.
    \param upper Upper bound.
    \param label Label to describe the value.
    """
    if value < lower:
        logging.warning(f"{label}={value} clamped to {lower=}")
        return lower
    elif value > upper:
        logging.warning(f"{label}={value} clamped to {upper=}")
        return upper
    return value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

## \namespace upkie.envs.upkie_ground_velocity
## \brief Ground-velocity Gymnasium environment.

import math
from typing import Dict, Optional, Tuple

import gymnasium as gym
import numpy as np

from upkie.exceptions import UpkieException
from upkie.utils.clamp import clamp_and_warn
from upkie.utils.filters import low_pass_filter
from upkie.utils.spdlog import logging

from .upkie_servos import UpkieServos


class UpkieGroundVelocity(gym.Wrapper):
    r"""!
    Environment where Upkie is used as a wheeled inverted pendulum.

    \anchor upkie_ground_velocity_description

    With this environment, Upkie keeps its legs straight and actions only
    affect wheel velocities. This way, it behaves like a <a
    href="https://scaron.info/robotics/wheeled-inverted-pendulum-model.html">wheeled
    inverted pendulum</a>. This ground-velocity environment is used for
    instance by the [MPC balancer](https://github.com/upkie/mpc_balancer/) and
    [PPO balancer](https://github.com/upkie/ppo_balancer) agents.

    \note For reinforcement learning with neural-network policies: the
    observation space and action space are not normalized.

    ### Action space

    The action corresponds to the ground velocity resulting from wheel
    velocities. The action vector is simply:

    \f[
    a =\begin{bmatrix} \dot{p}^* \end{bmatrix}
    \f]

    where we denote by \f$\dot{p}^*\f$ the commanded ground velocity in [m] /
    [s], which is internally converted into wheel velocity commands. Note that,
    while this action is not normalized, [-1, 1] m/s is a reasonable range for
    ground velocities.

    ### Observation space

    Vectorized observations have the following structure:

    \f[
    \begin{align*}
    o &= \begin{bmatrix} \theta \\ p \\ \dot{\theta} \\ \dot{p} \end{bmatrix}
    \end{align*}
    \f]

    where we denote by:

    - \f$\theta\f$ the pitch angle of the base with respect to the world
      vertical, in radians. This angle is positive when the robot leans
      forward.
    - \f$p\f$ the position of the average wheel contact point, in meters.
    - \f$\dot{\theta}\f$ the body angular velocity of the base frame along its
      lateral axis, in radians per seconds.
    - \f$\dot{p}\f$ the velocity of the average wheel contact point, in meters
      per seconds.

    As with all Upkie environments, full observations from the spine (detailed
    in \ref observations) are also available in the `info` dictionary
    returned by the reset and step functions.
    """

    ## \var action_space
    ## Action space.
    action_space: gym.spaces.Box

    ## \var env
    ## Internal \ref upkie.envs.upkie_servos.UpkieServos environment.
    env: UpkieServos

    ## \var fall_pitch
    ## Fall detection pitch angle, in radians.
    fall_pitch: float

    ## \var left_wheeled
    ## Set to True (default) if the robot is left wheeled, that is, a positive
    ## turn of the left wheel results in forward motion. Set to False for a
    ## right-wheeled variant.
    left_wheeled: bool

    ## \var observation_space
    ## Observation space.
    observation_space: gym.spaces.Box

    ## \var version
    ## Environment version number.
    version = 4

    ## \var wheel_radius
    ## Wheel radius in [m].
    wheel_radius: float

    def __init__(
        self,
        env: UpkieServos,
        fall_pitch: float = 1.0,
        left_wheeled: bool = True,
        max_ground_velocity: float = 1.0,
        wheel_radius: float = 0.06,
    ):
        r"""!
        Initialize environment.

        \param env UpkieServos environment to command servomotors.
        \param fall_pitch Fall detection pitch angle, in radians.
        \param left_wheeled Set to True (default) if the robot is left wheeled,
            that is, a positive turn of the left wheel results in forward
            motion. Set to False for a right-wheeled variant.
        \param max_ground_velocity Maximum commanded ground velocity in m/s.
            The default value of 1 m/s is conservative, don't hesitate to
            increase it once you feel confident in your agent.
        \param wheel_radius Wheel radius in [m].
        """
        super().__init__(env)
        if env.frequency is None:
            raise UpkieException("This environment needs a loop frequency")

        MAX_BASE_PITCH: float = np.pi
        MAX_GROUND_POSITION: float = float("inf")
        MAX_BASE_ANGULAR_VELOCITY: float = 1000.0  # rad/s
        observation_limit = np.array(
            [
                MAX_BASE_PITCH,
                MAX_GROUND_POSITION,
                MAX_BASE_ANGULAR_VELOCITY,
                max_ground_velocity,
            ],
            dtype=float,
        )
        action_limit = np.array([max_ground_velocity], dtype=float)

        # gymnasium.Env: observation_space
        self.observation_space = gym.spaces.Box(
            -observation_limit,
            +observation_limit,
            shape=observation_limit.shape,
            dtype=observation_limit.dtype,
        )

        # gymnasium.Env: action_space
        self.action_space = gym.spaces.Box(
            -action_limit,
            +action_limit,
            shape=action_limit.shape,
            dtype=action_limit.dtype,
        )

        # Class attributes
        self.__leg_servo_action = env.get_neutral_action()
        self.env = env
        self.fall_pitch = fall_pitch
        self.left_wheeled = left_wheeled
        self.wheel_radius = wheel_radius

    def __get_observation(self, spine_observation: dict) -> np.ndarray:
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Spine observation dictionary.
        \return Environment observation vector.
        """
        base_orientation = spine_observation["base_orientation"]
        pitch_base_in_world = base_orientation["pitch"]
        angular_velocity_base_in_base = base_orientation["angular_velocity"]
        ground_position = spine_observation["wheel_odometry"]["position"]
        ground_velocity = spine_observation["wheel_odometry"]["velocity"]

        obs = np.empty(4, dtype=float)
        obs[0] = pitch_base_in_world
        obs[1] = ground_position
        obs[2] = angular_velocity_base_in_base[1]
        obs[3] = ground_velocity
        return obs

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[np.ndarray, Dict]:
        r"""!
        Resets the environment and get an initial observation.

        \param seed Number used to initialize the environment’s internal random
            number generator.
        \param options Currently unused.
        \return
            - `observation`: Initial vectorized observation, i.e. an element
              of the environment's `observation_space`.
            - `info`: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        _, info = self.env.reset(seed=seed, options=options)
        spine_observation = info["spine_observation"]
        for joint in self.env.model.upper_leg_joints:
            position = spine_observation["servo"][joint.name]["position"]
            self.__leg_servo_action[joint.name]["position"] = position
        observation = self.__get_observation(spine_observation)
        return observation, info

    def __get_leg_servo_action(self) -> Dict[str, Dict[str, float]]:
        r"""!
        Get servo actions for both hip and knee joints.

        \return Servo action dictionary.
        """
        for joint in self.env.model.upper_leg_joints:
            prev_position = self.__leg_servo_action[joint.name]["position"]
            new_position = low_pass_filter(
                prev_output=prev_position,
                new_input=0.0,  # go to neutral configuration
                cutoff_period=1.0,  # in roughly one second
                dt=self.env.dt,
            )
            self.__leg_servo_action[joint.name]["position"] = new_position
        return self.__leg_servo_action

    def __get_wheel_servo_action(
        self, left_wheel_velocity: float
    ) -> Dict[str, Dict[str, float]]:
        r"""!
        Get servo actions for wheel joints.

        \param[in] left_wheel_velocity Left-wheel velocity, in [rad] / [s].
        \return Servo action dictionary.
        """
        right_wheel_velocity = -left_wheel_velocity
        servo_action = {
            "left_wheel": {
                "position": math.nan,
                "velocity": left_wheel_velocity,
            },
            "right_wheel": {
                "position": math.nan,
                "velocity": right_wheel_velocity,
            },
        }
        for joint in self.env.model.wheel_joints:
            servo_action[joint.name]["maximum_torque"] = joint.limit.effort
        return servo_action

    def __get_servo_action(self, action: np.ndarray) -> Dict[str, dict]:
        r"""!
        Convert environment action to a spine action dictionary.

        \param action Environment action.
        \return Spine action dictionary.
        """
        ground_velocity = clamp_and_warn(
            action[0],
            self.action_space.low[0],
            self.action_space.high[0],
            label="ground_velocity",
        )
        wheel_velocity = ground_velocity / self.wheel_radius
        left_wheel_sign = 1.0 if self.left_wheeled else -1.0
        left_wheel_velocity = left_wheel_sign * wheel_velocity
        leg_servo_action = self.__get_leg_servo_action()
        wheel_servo_action = self.__get_wheel_servo_action(left_wheel_velocity)
        return leg_servo_action | wheel_servo_action  # wheel comes second

    def __detect_fall(self, spine_observation: dict) -> bool:
        r"""!
        Detect a fall based on the base-to-world pitch angle.

        \param spine_observation Spine observation dictionary.
        \return True if and only if a fall is detected.

        Spine observations should have a "base_orientation" key. This requires
        the \ref upkie::cpp::observers::BaseOrientation observer in the spine's
        observer pipeline.
        """
        pitch = spine_observation["base_orientation"]["pitch"]
        if abs(pitch) > self.fall_pitch:
            logging.warning(
                "Fall detected (pitch=%.2f rad, fall_pitch=%.2f rad)",
                abs(pitch),
                self.fall_pitch,
            )
            return True
        return False

    def step(
        self,
        action: np.ndarray,
    ) -> Tuple[np.ndarray, float, bool, bool, dict]:
        r"""!
        Run one timestep of the environment's dynamics.

        When the end of the episode is reached, you are responsible for calling
        `reset()` to reset the environment's state.

        \param action Action from the agent.
        \return
            - `observation`: Observation of the environment, i.e. an element
              of its `observation_space`.
            - `reward`: Reward returned after taking the action.
            - `terminated`: Whether the agent reached a terminal state,
              which may be a good or a bad thing. When true, the user needs to
              call `reset()`.
            - `truncated`: Whether the episode is reaching max number of
              steps. This boolean can signal a premature end of the episode,
              i.e. before a terminal state is reached. When true, the user
              needs to call `reset()`.
            - `info`: Dictionary with additional information, reporting in
              particular the full observation dictionary coming from the spine.
        """
        servo_action = self.__get_servo_action(action)
        _, reward, terminated, truncated, info = self.env.step(servo_action)
        spine_observation = info["spine_observation"]
        observation = self.__get_observation(spine_observation)
        if self.__detect_fall(spine_observation):
            terminated = True
        return observation, reward, terminated, truncated, info


def make_upkie_ground_velocity(**kwargs):
    r"""!
    Make a new ground-velocity environment.

    This function is meant to be called by `gymnasium.make()` rather than to be
    called directly.

    \param kwargs Keyword arguments forwarded to both the \ref
        upkie.envs.upkie_ground_velocity.UpkieGroundVelocity and internal \ref
        upkie.envs.upkie_servos.UpkieServos environments.
    \return New ground-velocity environment.
    """
    ground_velocity_kwargs = {
        key: value
        for key, value in kwargs.items()
        if key
        in (
            "fall_pitch",
            "left_wheeled",
            "max_ground_velocity",
            "wheel_radius",
        )
    }
    servos_kwargs = {
        key: value
        for key, value in kwargs.items()
        if key not in ground_velocity_kwargs
    }
    env = UpkieServos(**servos_kwargs)
    return UpkieGroundVelocity(env, **ground_velocity_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria


class JointLimit:
    """!
    Joint limit in configuration and tangent spaces.

    Attributes of this class match the URDF specification.
    """

    ## \var lower
    ## Lower configuration limit of the joint, in radians.
    lower: float

    ## \var upper
    ## Upper configuration limit of the joint, in radians.
    upper: float

    ## \var velocity
    ## Velocity limit of the joint, in [rad] / [s].
    velocity: float

    ## \var effort
    ## Torque limit of the joint, in [N m].
    effort: float

    def __init__(
        self,
        lower: float,
        upper: float,
        velocity: float,
        effort: float,
    ):
        """!
        Manual constructor for joint limits.

        (We don't use a `dataclass` because Doxygen does not detect
        attribute-only declarations in Python as of version 1.9.1.)
        """
        self.lower = lower
        self.upper = upper
        self.velocity = velocity
        self.effort = effort



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""!
Functions to work with nested dictionaries.
"""


def nested_update(target_dict: dict, new_dict: dict) -> None:
    r"""!
    Update a target dictionary recursively.

    Consider the following example:

        >>> d = {"a": {"b": 1}}
        >>> d2 = {"a": {"c": 2}}
        >>> d.update(d2)
        >>> d
        {'a': {'c': 2}}

    With this function:

        >>> d = {"a": {"b": 1}}
        >>> d2 = {"a": {"c": 2}}
        >>> d.update(d2)
        >>> d
        {'a': {'b': 1, 'c': 2}}

    \param target_dict Output dictionary.
    \param new_dict Input dictionary with new values to update.
    """
    for key, value in new_dict.items():
        if (
            key in target_dict
            and isinstance(target_dict[key], dict)
            and isinstance(value, dict)
        ):
            nested_update(target_dict[key], value)
        else:
            target_dict[key] = new_dict[key]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

import numpy as np


class WheeledInvertedPendulumReward:
    r"""!
    Weights of the position and velocity terms in rewards.
    """

    ## \var position_weight
    ## Weight of the position term.
    position_weight: float

    ## \var velocity_weight
    ## Weight of the velocity term.
    velocity_weight: float

    def __init__(
        self, position_weight: float = 1.0, velocity_weight: float = 1.0
    ):
        r"""!
        Initialize with reward weights.

        \param position Weight of the position term.
        \param velocity Weight of the velocity term.
        """
        self.position_weight = position_weight
        self.velocity_weight = velocity_weight

    def __call__(
        self,
        pitch: float,
        ground_position: float,
        angular_velocity: float,
        ground_velocity: float,
    ) -> float:
        r"""!
        Get reward for a given state.

        \param[in] pitch Pitch angle of the rotation from base to world, in
            [rad].
        \param[in] ground_position Position on the ground, in [m].
        \param[in] angular_velocity Angular velocity of the rotation from base
            to world, in [rad] / [s].
        \p[aram[in] ground_velocity Ground velocity, in [m] / [s].
        \return Reward.
        """
        tip_height = 0.58  # [m]
        tip_position = ground_position + tip_height * np.sin(pitch)
        tip_velocity = (
            ground_velocity + tip_height * angular_velocity * np.cos(pitch)
        )

        std_position = 0.05  # [m]
        position_reward = np.exp(-((tip_position / std_position) ** 2))
        velocity_penalty = -abs(tip_velocity)

        return (
            self.position_weight * position_reward
            + self.velocity_weight * velocity_penalty
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""!
Serialization function.
"""


def serialize(obj):
    r"""!
    Serialize an object for message packing.

    \param obj Object to serialize.

    \return Serialized object.

    \note Calling the numpy conversion is much faster than the default list
    constructor:

    \code{python}
    In [1]: x = random.random(6)

    In [2]: %timeit list(x)
    789 ns ± 5.79 ns per loop (mean ± std. dev. of 7 runs, 1e6 loops each)

    In [3]: %timeit x.tolist()
    117 ns ± 0.865 ns per loop (mean ± std. dev. of 7 runs, 1e7 loops each)
    \endcode
    """
    if hasattr(obj, "tolist"):  # numpy.ndarray
        return obj.tolist()
    elif hasattr(obj, "np"):  # pinocchio.SE3
        return obj.np.tolist()
    elif hasattr(obj, "serialize"):  # more complex objects
        return obj.serialize()
    return obj



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

from typing import List, Tuple
from xml.etree import ElementTree

import numpy as np

from .joint import Joint
from .joint_limit import JointLimit


class Model:
    """!
    Robot model parsed from its URDF description.
    """

    ## \var joints
    ## Joints of the robot model.
    joints: List[Joint]

    ## \var rotation_ars_to_world
    ## Rotation matrix from the ARS frame to the world frame.
    rotation_ars_to_world: np.ndarray

    ## \var rotation_base_to_imu
    ## Rotation matrix from the base frame to the IMU frame.
    rotation_base_to_imu: np.ndarray

    ## \var upper_leg_joints
    ## Upper-leg (hip and knee) joints.
    upper_leg_joints: Tuple[float]

    ## \var wheel_joints
    ## Wheel joints.
    wheel_joints: Tuple[float]

    def __init__(self, urdf_path: str):
        r"""!
        Constructor for the robot model wrapper.

        \param[in] urdf_path Path to the robot description.
        """
        tree = ElementTree.parse(urdf_path)
        joint_tags = [
            child for child in tree.getroot() if child.tag == "joint"
        ]
        limits = [
            (joint, child)
            for joint in joint_tags
            for child in joint
            if child.tag == "limit"
        ]
        joints = []
        for idx, (joint, limit) in enumerate(limits):
            joint_name = joint.attrib["name"]
            joints.append(
                Joint(
                    index=idx + 1,  # starts from 1
                    idx_q=idx,
                    idx_v=idx,
                    name=joint_name,
                    limit=JointLimit(
                        lower=float(limit.attrib.get("lower", -np.inf)),
                        upper=float(limit.attrib.get("upper", +np.inf)),
                        effort=float(limit.attrib["effort"]),
                        velocity=float(limit.attrib["velocity"]),
                    ),
                )
            )
        upper_leg_joints = tuple(
            joint
            for joint in joints
            if joint.name
            in ("left_hip", "left_knee", "right_hip", "right_knee")
        )
        wheel_joints = tuple(
            joint
            for joint in joints
            if joint.name in ("left_wheel", "right_wheel")
        )
        self.joints = joints
        self.rotation_ars_to_world = np.diag([1.0, -1.0, -1.0])
        self.rotation_base_to_imu = np.diag([-1.0, 1.0, -1.0])
        self.upper_leg_joints = upper_leg_joints
        self.wheel_joints = wheel_joints



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

import gymnasium as gym
import numpy as np

from upkie.exceptions import UpkieException


class RandomPush(gym.Wrapper):
    """!
    At each step, with a given probability, apply a random push to the robot.
    """

    ## \var env
    ## Wrapped environment.
    env: gym.Env

    ## \var push_generator
    ## Function that generates the push force. It should return a 3D numpy
    ## array.
    push_generator: callable

    ## \var push_prob
    ## Probability of pushing at each step.
    push_prob: float

    def __init__(
        self,
        env: gym.Env,
        push_prob: float = 0.01,
        push_generator: callable = lambda: np.random.normal(0, 400, 3),
    ):
        r"""!
        Initialize wrapper.

        \param env Environment to wrap.
        \param push_prob Probability of applying a push at each timestep.
        \param push_generator Function that generates the push force. It should
            return a 3D NumPy array.
        """
        super().__init__(env)
        if not env.has_wrapper_attr("set_bullet_action"):
            raise UpkieException(
                "Wrapped environment must have a `set_bullet_action` method"
            )
        self.env = env
        self.push_prob = push_prob
        self.push_generator = push_generator

    def step(self, action):
        """Adds a random push to the action."""
        if np.random.binomial(1, self.push_prob):
            force = self.push_generator()
            self.env.unwrapped.set_bullet_action(
                {"external_forces": {"torso": {"force": force}}}
            )
        else:  # Reset the forces to zero
            self.env.unwrapped.set_bullet_action(
                {"external_forces": {"torso": {"force": np.zeros(3)}}}
            )
        return self.env.step(action)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""!
Basic digital filters.
"""

from typing import Tuple

from .clamp import clamp


def abs_bounded_derivative_filter(
    prev_output: float,
    new_input: float,
    dt: float,
    max_output: float,
    max_derivative: float,
) -> float:
    r"""!
    Filter signal so that the absolute values of its output and output
    derivative stay within bounds.

    \param prev_output Previous filter output, or initial value.
    \param new_input New filter input.
    \param dt Sampling period in [s].
    \param max_output Maximum absolute value of the output.
    \param max_derivative Maximum absolute value of the output derivative.
    \return New filter output.
    """
    return bounded_derivative_filter(
        prev_output,
        new_input,
        dt,
        (-max_output, max_output),
        (-max_derivative, max_derivative),
    )


def bounded_derivative_filter(
    prev_output: float,
    new_input: float,
    dt: float,
    output_bounds: Tuple[float, float],
    derivative_bounds: Tuple[float, float],
) -> float:
    r"""!
    Filter signal so that its output and output derivative stay within bounds.

    \param prev_output Previous filter output, or initial value.
    \param new_input New filter input.
    \param dt Sampling period in [s].
    \param output_bounds Min and max value for the output.
    \param derivative_bounds Min and max value for the output derivative.
    \return New filter output.
    """
    derivative = (new_input - prev_output) / dt
    derivative = clamp(derivative, *derivative_bounds)
    output = prev_output + derivative * dt
    return clamp(output, *output_bounds)


def low_pass_filter(
    prev_output: float,
    cutoff_period: float,
    new_input: float,
    dt: float,
) -> float:
    r"""!
    Low-pass filter.

    \param prev_output Previous filter output, or initial value.
    \param cutoff_period Time constant of the filter in [s].
    \param new_input New filter input.
    \param dt Sampling period in [s].
    \return New filter output.
    """
    alpha = dt / cutoff_period
    assert alpha < 0.5  # Nyquist-Shannon sampling theorem
    return prev_output + alpha * (new_input - prev_output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""!
Functions to work on the onboard Raspberry Pi.
"""

import os
import sys

from upkie.exceptions import UpkieRuntimeError

from .spdlog import logging

__AGENT_CPUID: int = 3
__MODEL_PATH: str = "/sys/firmware/devicetree/base/model"
__ON_RASPI: bool = False

if os.path.exists(__MODEL_PATH):
    with open(__MODEL_PATH, "r", encoding="utf-8") as fh:
        line = fh.readline()
        __ON_RASPI = line.startswith("Raspberry")


def on_raspi() -> bool:
    """!
    Check whether we are running on a Raspberry Pi.
    """
    return __ON_RASPI


def configure_agent_process() -> None:
    r"""!
    Configure process to run as an agent on the Raspberry Pi.

    \note This function assumes we are running an underlying script. It won't
    work from an interpreter.
    """
    if hasattr(sys, "ps1"):
        raise UpkieRuntimeError(
            "Cannot configure agent process from an interpreter"
        )
    if os.geteuid() != 0:
        logging.info("Re-running as root to set the CPU affinity")
        args = ["sudo", "-E", sys.executable] + sys.argv + [os.environ]
        os.execlpe("sudo", *args)
    os.sched_setaffinity(0, {__AGENT_CPUID})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""!
State-machine request identifiers.
"""

from enum import IntEnum


class Request(IntEnum):
    r"""!
    Flag set by the agent to request an operation from the spine.

    Once the spine has processed the request, it resets the flag to \ref kNone
    in shared memory.

    @var kNone
    Flag set when there is no active request.

    @var kAction
    Flag set to indicate an action has been supplied.

    @var kStart
    Flag set to start the spine.

    @var kStop
    Flag set to stop the spine.

    @var kError
    Flag set when the last request was invalid.
    """

    kNone = 0
    kAction = 1
    kStart = 2
    kStop = 3
    kError = 4



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from .add_action_to_observation import AddActionToObservation
from .add_lag_to_action import AddLagToAction
from .differentiate_action import DifferentiateAction
from .noisify_action import NoisifyAction
from .noisify_observation import NoisifyObservation
from .observation_based_reward import ObservationBasedReward
from .random_push import RandomPush

__all__ = [
    "AddActionToObservation",
    "AddLagToAction",
    "DifferentiateAction",
    "NoisifyAction",
    "NoisifyObservation",
    "ObservationBasedReward",
    "RandomPush",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Tuple, Union

import gymnasium as gym
import numpy as np

from upkie.utils.filters import low_pass_filter


class AddLagToAction(gym.Wrapper):
    """!
    Model lag by applying a low-pass filter to the action of an environment.

    Note that there is a difference between "delay" and "lag":

    - Delay is a fixed time interval corresponding to the time it takes the
      input to affect the output, e.g. `output(t + delay) = input(t)`.
    - Lag is a phase shift in the system's response. It can be thought of as a
      gradual response to input changes where the output does not immediately
      match the input, e.g. `d/dt{output}(t) = lag * (input(t) - output(t))`.

    In this wrapper, we model *lag* where the output (action forward to the
    wrapped environment) is a low-pass filtered version of the input (action
    passed to `step`).
    """

    ## \var filtered_action
    ## Wrapped action after low-pass filtering.
    filtered_action: np.ndarray

    ## \var time_constant
    ## Cutoff period in seconds of a low-pass filter applied to the action.
    time_constant: float

    ## \var time_constant_box
    ## Box space from which the time constant is sampled at every reset of the
    ## environment.
    time_constant_box: gym.spaces.Box

    def __init__(self, env, time_constant: Union[float, gym.spaces.Box]):
        r"""!
        Initialize wrapper.

        \param env Environment to wrap.
        \param time_constant Cutoff period in seconds of the low-pass filter
            applied to the action. If a Box is provided, couple of lower and
            upper bounds for the action: a new time constant is sampled
            uniformly at random between these bounds at every reset of the
            environment.
        """
        super().__init__(env)
        time_constant_box = (
            time_constant
            if isinstance(time_constant, gym.spaces.Box)
            else gym.spaces.Box(
                low=time_constant - 1e-10, high=time_constant + 1e-10
            )
        )
        self.filtered_action = np.zeros(env.action_space.shape)
        self.time_constant = time_constant_box.sample()
        self.time_constant_box = time_constant_box

    def reset(self, **kwargs):
        r"""!
        Reset the environment.

        \param kwargs Keyword arguments forwarded to the wrapped environment.
        """
        self.filtered_action = np.zeros(self.env.action_space.shape)
        self.time_constant = self.time_constant_box.sample()
        return self.env.reset(**kwargs)

    def step(
        self, action: np.ndarray
    ) -> Tuple[np.ndarray, float, bool, bool, dict]:
        r"""!
        Step the environment.

        \param action Action from the agent.
        \return Tuple with (observation, reward, terminated, truncated,info).
            See \ref upkie.envs.upkie_servos.UpkieServos.step for details.
        """
        dt = self.env.unwrapped.dt
        if self.time_constant <= 2.0 * dt:
            # Nyquist–Shannon sampling theorem
            return self.env.step(action)

        self.filtered_action = low_pass_filter(
            self.filtered_action,
            self.time_constant,
            action,
            dt,
        )
        return self.env.step(self.filtered_action)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

import logging
import time
from multiprocessing import resource_tracker
from multiprocessing.shared_memory import SharedMemory

from ..exceptions import SpineError


def wait_for_shared_memory(
    shm_name: str,
    retries: int,
) -> SharedMemory:
    r"""!
    Connect to the spine shared memory.

    \param shm_name Name of the shared memory object.
    \param retries Number of times to try opening the shared-memory file.
    \throw SpineError If the spine did not respond after the prescribed number
        of trials.
    """
    # Remove leading slash if present, as SharedMemory will prepend it
    # See https://github.com/upkie/upkie/issues/375
    shm_name = shm_name.lstrip("/")

    for trial in range(retries):
        if trial > 0:
            logging.info(
                f"Waiting for spine /{shm_name} to start "
                f"(trial {trial} / {retries})..."
            )
            time.sleep(1.0)
        try:
            shared_memory = SharedMemory(shm_name, size=0, create=False)
            # Why we unregister: https://github.com/upkie/upkie/issues/376
            # Upstream issue: https://github.com/python/cpython/issues/82300
            resource_tracker.unregister(shared_memory._name, "shared_memory")
            return shared_memory
        except FileNotFoundError:
            pass
    raise SpineError(
        f"spine /{shm_name} did not respond after {retries} attempts"
    )
